# Description
#image should Be
#1-centered in the image.
#2-rotated that such the eyes lie on a horizontal line (i.e., the face is rotated such that the eyes lie along the same y-coordinates).
#3-scaled such that the size of the faces are approximately identical.

#align face is a funtion depend on landmark(68) and dlib image extraction

from imutils.face_utils import FaceAligner
from imutils.face_utils import rect_to_bb
import argparse
import imutils
import dlib
#import cv2
#import uuid


#image = cv2.imread("input.png")
#f = str(uuid.uuid4())


#send face rectangle
def align_faces(image):
    #initialize dlib detector && landmarks Predictor && FaceAligner class
    detector = dlib.get_frontal_face_detector()
    pred = "shape_predictor_68_face_landmarks.dat"
    predictor = dlib.shape_predictor(pred)
    fa = FaceAligner(predictor, desiredFaceWidth=256)

    image = imutils.resize(image, width=800)
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    #detect faces in gray scale
    rects = detector(gray,2)
    for rect in rects:
        (x, y, w, h) = rect_to_bb(rect)
        faceOrig = imutils.resize(image[y:y + h, x:x + w], width=256)
        faceAligned = fa.align(image, gray, rect)

    #to display the output images
    # cv2.imwrite("foo/" + f + ".png", faceAligned)
    # cv2.imshow("Original", faceOrig)
    # cv2.imshow("Aligned", faceAligned)
    # cv2.waitKey(0)
    return faceAligned



import numpy as np
import cv2
import dlib

__dlib_detector = dlib.get_frontal_face_detector()

# public interface
def get_faces(img):
	rects = __dlib_detector(img, 2)
	res = []
	for rect in rects:
		p1 = rect.tl_corner()
		p2 = rect.br_corner()
		if (p1.x - p2.x == 0) or (p1.y - p2.y == 0): continue
		face = _extract_face(img, rect)
		face = _enhance_face(face)
		box = [(p1.x, p1.y), (p2.x, p2.y)]
		res.append((face, box))
	return res

def is_one_face(img):
	return len(__dlib_detector(img, 2)) == 1

# private methods
def _extract_face(img, rect):
	p1 = rect.tl_corner()
	p2 = rect.br_corner()
	face = img[p1.y:p2.y, p1.x:p2.x]
	return face

def _enhance_face(face):
	return face
	face = cv2.resize(face, (48, 48))
	face = cv2.cvtColor(face, cv2.COLOR_RGB2GRAY)
	return face


if __name__ == '__main__':
	im = cv2.imread('y.jpg')
	cv2.imshow("title", im)
	cv2.waitKey(0)
	res = get_faces(im)
	for i in res:
		cv2.imshow("df", i[0])
		cv2.waitKey(0)



#!usr/env/bin python3

import cv2
import numpy as np
from moviepy.editor import *
import ffmpeg 

from Cerebro.interface import process_image as pi
from Cerebro.image import FaceTracking as tracker
from Cerebro.model.emotions_model import EmotionsModel

from Cerebro.image.face_detector import detect_dlib
from Cerebro.image.face_detector import detect_haar
from Cerebro.image.face_detector import detect_lbp

model = EmotionsModel()

def check_rotation(path_video_file):

    # this returns meta-data of the video file in form of a dictionary
    meta_dict = ffmpeg.probe(path_video_file)
    #print(meta_dict['streams'][0]['tags'])

    # from the dictionary, meta_dict['streams'][0]['tags']['rotate'] is the key
    # we are looking for
    rotateCode = None
    if 'rotate' in meta_dict['streams'][0]['tags']:
        if int(meta_dict['streams'][0]['tags']['rotate']) == 90:
            rotateCode = cv2.ROTATE_90_CLOCKWISE
        elif int(meta_dict['streams'][0]['tags']['rotate']) == 180:
            rotateCode = cv2.ROTATE_180
        elif int(meta_dict['streams'][0]['tags']['rotate']) == 270:
            rotateCode = cv2.ROTATE_90_COUNTERCLOCKWISE

    return rotateCode

def detect_video_emotions_with_tracking (video_path, output_path, batch_size=125, detector_type='dlib', verbose=False):
    #saving audio 
    video = VideoFileClip(video_path)
    audio = video.audio

    #detector
    if detector_type == 'dlib':
        detector = detect_dlib
    elif detector_type =='haar':
        detector = detect_haar
    elif detector_type =='lbp':
        detector = detect_lbp

    #processing frames
    vidObj = cv2.VideoCapture(video_path)
    fps = vidObj.get(5) #fps
    rotateCode = check_rotation(video_path)

    success = 1 #checks whether frames were extracted
    it = 0
    img_frames = []
    to_be_tracked = []
    to_be_printed = 0
    #getting frames
    while success:
        success, image = vidObj.read()

        if rotateCode is not None:
            image = cv2.rotate(image, rotateCode)

        if not success :
            break

        #after calling track function that is what is supposed to be done
        if it == batch_size :
            returned_faces, returned_cords = tracker.faceTracking(to_be_tracked, batch_size)
            
            #print(len(returned_faces))

            emotions = model.predict_with_vote(returned_faces)
            
            if len(returned_cords) == 0:
                for i in range(batch_size):
                    tmp = to_be_tracked[i][0]
                    tmp = cv2.cvtColor(tmp,cv2.COLOR_BGR2RGB)
                    img_frames.append(tmp)
            else:
                for j in range(len(returned_cords[0])): 
                    tmp = to_be_tracked[j][0]
                    extracted_faces_emotions = []
                    for i in range(len(returned_cords)):
                        if returned_cords[i][j] != None:
                            em = emotions[i]
                            extracted_faces_emotions.append((returned_faces[i][j], returned_cords[i][j], em))
                            
                    tmp = pi.mark_faces_emotions(tmp, extracted_faces_emotions=extracted_faces_emotions)
                    tmp = cv2.cvtColor(tmp,cv2.COLOR_BGR2RGB)
                    img_frames.append(tmp)
            
            to_be_tracked = []
            it = 0
            to_be_printed += 1
            if verbose:
                print("batch: ", to_be_printed)


        face_boxes = []
        frame_data = detector.get_faces(image)
        for fd in frame_data :
            face_boxes.append(fd[1])
        to_be_tracked.append([image, face_boxes])
        it += 1

    #making output
    clips = [ImageClip(m).set_duration(1/fps) for m in img_frames]
    concat_clip = concatenate_videoclips(clips, method="chain")
    concat_clip_edited = concat_clip.set_audio(audio)
    concat_clip_edited.write_videofile(output_path, fps=fps)

def detect_video_emotions(video_path, output_path, skip = 50, detector_type = 'dlib', verbose=False):
    """
    skip >> determines number of skipped frames during procerssing
    in normal video speed skip=50 means detect emotions every 2 seconds
    higher skip means faster processing but less accurate
    -----------------------------------------
    output_path >> include output name & extension
    the output video extension should be mp4
    """
    #saving audio 
    video = VideoFileClip(video_path)
    audio = video.audio

    #processing frames
    vidObj = cv2.VideoCapture(video_path)
    fps = vidObj.get(5) #fps
    rotateCode = check_rotation(video_path)

    success = 1 #checks whether frames were extracted
    real_frame_counter = 1 #to check with sampling
    sampled_frame_counter = 1 #to print frame number for user
    prev_frame_data = None

    img_frames = []

    #getting frames
    while success:
        success, image = vidObj.read()

        if rotateCode is not None:
            image = cv2.rotate(image, rotateCode)

        if not success :
            break
        #processing frames
        if not(real_frame_counter % skip):
            prev_frame_data = pi.extract_faces_emotions(image, detector_type)
            if verbose: print("process frame: ", sampled_frame_counter)
            sampled_frame_counter += 1

        image = pi.mark_faces_emotions(image,None, prev_frame_data)
        real_frame_counter += 1
        image = cv2.cvtColor(image,cv2.COLOR_BGR2RGB)
        img_frames.append(image)
    
    #making output video
    if verbose: print("making output video...")
    clips = [ImageClip(m).set_duration(1/fps) for m in img_frames]
    concat_clip = concatenate_videoclips(clips, method="chain")
    concat_clip_edited = concat_clip.set_audio(audio)
    concat_clip_edited.write_videofile(output_path, fps=fps)



#! /usr/bin/env python3

import numpy as np
import cv2
# import matplotlib.pyplot as plt

import Cerebro.image.face_detector.detect_dlib as dlib_detector
import Cerebro.image.face_detector.detect_haar as haar_detector
import Cerebro.image.face_detector.detect_lbp as lbp_detector

def get_squared(img, n):
    ret = img.copy()
    width = height = n # np.max((img.shape[0], img.shape[1]))
    ret = cv2.resize(img, (width, height),interpolation = cv2.INTER_AREA)
    return ret

def extend(img):
    ret = img.copy()
    if img.shape[0] < img.shape[1]:
        diff = img.shape[1] - img.shape[0]
        add = np.array([[[0] * 3] * img.shape[1]] * diff)
        ret = np.vstack((add, img))
    if img.shape[0] > img.shape[1]:
        diff = img.shape[0] - img.shape[1]
        add = np.array([[[0] * 3] * diff] * img.shape[0])
        ret = np.hstack((add, img))
    return ret

def normalize_image(img, n, detect=False, detector='dlib'):
    assert isinstance(img, np.ndarray), "image must be of type np.ndarray"
    assert isinstance(n, int), "n must be integer"
    # print(img.shape)
    ret = img.copy()

    # apply face detection if asked
    if detect:
        if detector == 'dlib':
            ret = dlib_detector.get_faces(img)[0][0]
        elif detector == 'haar':
            ret = haar_detector.get_faces(img)[0][0]
        elif detector == 'lbp':
            ret = lbp_detector.get_faces(img)[0][0]
        else:
            raise Exception('unknown detector %s given' %detector)

    # making the image a square image by adding padding depending on the shape
    if ret.shape[0] < ret.shape[1]:
        diff = ret.shape[1] - ret.shape[0]
        ret = cv2.copyMakeBorder(ret,diff,0,0,0,cv2.BORDER_CONSTANT,value=[0,0,0])
    elif ret.shape[0] > ret.shape[1]:
        diff = ret.shape[0] - ret.shape[1]
        ret = cv2.copyMakeBorder(ret,0,diff,0,0,cv2.BORDER_CONSTANT,value=[0,0,0])

    # resize image to given size
    if ret.shape[0] != n:
        ret = cv2.resize(ret, (n, n))

    return ret

def normalize_channels(image):
    if len(image.shape) == 3 and image.shape[-1] == 3:
        ret = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    else:
        ret = image.copy()
    return ret



#! /user/bin/env python3

import os
os.environ['KERAS_BACKEND'] = 'theano'

import keras
from keras.models import Model
from keras.utils import to_categorical
# from keras.preprocessing.image import ImageDataGenerator
from keras.layers import Input, Conv2D, MaxPooling2D, Flatten, Dense, BatchNormalization, concatenate, Dropout

import pickle
import numpy as np
import os
import Cerebro
from pathlib import Path
from Cerebro.image.enhancement import filters
from Cerebro.image import feature_extraction, utils

class EmotionsModel(object):

    def __init__(
        self,
        verbose=False,
        create_new=False,
        use_reduced_emotions=False,
        use_hog=None,
        use_cnn=None,
        use_lm=None,
        emotions=None,
        reduced_emotions=None,
    ):
        # pathes constants
        self.model_path = os.path.join(Cerebro.__cwd__,'saved-models/emotions_model.f5')
        self.model_specs_path = os.path.join(Cerebro.__cwd__,'saved-models/emotions_model_specs.bin')

        if not os.path.isfile(self.model_path):
            raise Exception("emotions_model.f5 doesn't exist.") 

        if not os.path.isfile(self.model_specs_path):
            raise Exception("emotions_model_specs.bin doesn't exist.") 

        self.verbose = verbose
        self.use_reduced_emotions = use_reduced_emotions
        
        # model numbers
        self.imageSize = 150
        self.batch_size = 128
        self.epochs = 20


        if not create_new and self.has_saved_model():
            self.load_model()
            self.is_trained = True
            self.model._make_predict_function()
        else:
            if use_hog == None or use_cnn == None or use_lm == None or emotions == None or reduced_emotions == None:
                raise Exception(
                    'When creating new model, all model specs (use_hog, use_cnn, use_lm, emotions) must be given'
                )

            # set model specs
            self.use_hog = use_hog
            self.use_cnn = use_cnn
            self.use_lm = use_lm
            self.emotions = emotions
            self.reduced_emotions = reduced_emotions

            self.model = self.__create_model__()
            self.is_trained = False

    def fit(self, xs, ys, should_save_model=True, epochs_num=None):
        if epochs_num == None:
            epochs_num = self.epochs

        xs = self.__transform_input__(xs)
        ys = to_categorical(ys, len(self.emotions))

        # data_gen = ImageDataGenerator(horizontal_flip =True)
       	# history = self.model.fit_generator(data_gen.flow(max(xs),ys,self.batch_size),
       	# 	steps_per_epoch = len(xs)/self.batch_size,epochs=self.epochs)
		
        history = self.model.fit(
            xs, ys,
            batch_size=self.batch_size,
            epochs=self.epochs,
            validation_split=0.1,
            verbose=self.verbose,
        )

        self.is_trained = True
        if should_save_model:
            self.save_model()

        return history

    def test(self, faces, targets):
        if not self.is_trained:
            raise Exception("Model not trained yet")

        faces = self.__transform_input__(faces)
        targets = to_categorical(targets, len(self.emotions))
        return self.model.evaluate(faces, targets)

    def predict(self, faces, prob_emotion=False):
        if not self.is_trained:
            raise Exception("Model not trained yet")

        is_one_face = False
        if type(faces) is not list:
            faces = [faces]
            is_one_face = True

        faces = self.__transform_input__(faces)

        res = self.model.predict(faces, batch_size=self.batch_size).tolist()

        if not prob_emotion:
            for i, all in enumerate(res):
                if self.use_reduced_emotions:
                    res[i] = self.reduced_emotions[np.argmax(all)]
                else:
                    res[i] = self.emotions[np.argmax(all)]

        if is_one_face:
            res = res[0]

        return res
    
    # each element should be an array of the same face
    # returns vector of the same size, each dimension contains the resulting emotion from the vote
    def predict_with_vote(self, faces):

        if type(faces) is not list:
            faces = [faces]
            
        is_one_vector = False
        if faces:
            if type(faces[0]) is not list:
                faces = [faces]
                is_one_vector = True
        
        result = []
        
        for vector in faces:
            emotions_map = {}
            transformed_vector = [face for face in vector if type(face) != type(None)]
            emotions_vector = self.predict(transformed_vector)
            for emotion in emotions_vector:
                emotions_map[emotion] = emotions_map.get(emotion, 0) + 1
            sortedEmotions = sorted(emotions_map.items(), key=lambda x: x[1], reverse=True)
            result.append(sortedEmotions[0][0])
        
        if is_one_vector:
            result = result[0]
        
        return result

    def __transform_input__(self, images, should_enhance=True):
        lms, hogs, imgs = [], [], []

        if self.verbose:
            size = 30
            print()

        for i, image in enumerate(images):
            img = self.__enhance_image__(image) if should_enhance else image
            if self.use_cnn:
                imgs.append(np.reshape(img, (self.imageSize, self.imageSize, 1)))
            if self.use_lm:
                lms.append(feature_extraction.get_face_landmarks(img))
            if self.use_hog:
                hogs.append(feature_extraction.sk_get_hog(img, pixels_per_cell=(20, 20)))

            # printing the done percentage
            if self.verbose:
                done = (i+1) / len(images)
                print('\rExtracting Features - %3d%% [' % (done*100), end='')
                for j in range(size):
                    cnt = int(done*size)
                    if j <= cnt:
                        print('=', end='')
                    elif j == cnt+1:
                        print('>', end='')
                    else:
                        print('.', end='')
                print(']', end='')
        if self.verbose:
            print()

        ret = []
        if self.use_cnn:
            ret.append(np.array(imgs))
        if self.use_hog:
            ret.append(np.array(hogs))
        if self.use_lm:
            ret.append(np.array(lms))

        return ret

    def __enhance_image__(self, img):
        # normalize image to wanted size
        img = utils.normalize_image(img, self.imageSize)

        # make sure image is always in grayscale
        img = utils.normalize_channels(img)

        # remove salt and pepper
        img = filters.median(img)

        # sharpen images
        img = filters.laplacian(img)

        return img

    def save_model(self):
        # save model weights
        self.model.save(self.model_path)

        # save model specs
        with open(self.model_specs_path, 'wb') as specsFile:
            pickle.dump(
                (
                    self.use_hog,
                    self.use_cnn,
                    self.use_lm,
                    self.emotions,
                    self.reduced_emotions,
                ),
                specsFile
            )

    def load_model(self):
        # load model weights
        self.model = keras.models.load_model(self.model_path)

        # load model specs
        with open(self.model_specs_path, 'rb') as specsFile:
            self.use_hog, self.use_cnn, self.use_lm, self.emotions, self.reduced_emotions = pickle.load(specsFile)

    def has_saved_model(self):
        model_path = Path(self.model_path)
        model_specs_path = Path(self.model_specs_path)
        return model_path.is_file() and model_specs_path.is_file()

    def __create_model__(self):
        conv_activation = 'sigmoid'
        dense_activation = 'sigmoid'
        keep_prob = 0.956

        # ========================== CNN Part ==========================
        if self.use_cnn:
            input_image = Input(batch_shape=(None, 150, 150, 1), dtype='float32', name='input_image')
            x = BatchNormalization(axis=-1)(input_image)

            x = Conv2D(filters=8, kernel_size=(15, 15), padding='valid', activation=conv_activation)(input_image)
            x = BatchNormalization(axis=-1)(x)

            x = Conv2D(filters=16, kernel_size=(15, 15), padding='valid', activation=conv_activation)(x)
            x = BatchNormalization(axis=-1)(x)

            x = MaxPooling2D(pool_size=(2, 2), strides=2, data_format="channels_last")(x)

            x = Conv2D(filters=16, kernel_size=(15, 15), padding='valid', activation=conv_activation)(x)
            x = MaxPooling2D(pool_size=(2, 2), strides=2, data_format="channels_last")(x)
            x = BatchNormalization(axis=-1)(x)

            x = Flatten()(x)
            x = Dense(units=1024, activation=dense_activation)(x)
            x = Dropout(rate=keep_prob)(x)
            x = Dense(units=512, activation=dense_activation)(x)
            x = Dropout(rate=keep_prob)(x)

            x = BatchNormalization(axis=-1)(x)

            outputCNN = x

        # ========================== Image features part ==========================
        if self.use_lm or self.use_hog:

            if self.use_lm:
                inputLandmarks = Input(batch_shape=(None, 68, 2), dtype='float32', name='input_landmarks')
                flatLandmarks = Flatten()(inputLandmarks)
                normalizedLandmarks = BatchNormalization(axis=-1)(flatLandmarks)
                outputImage = normalizedLandmarks

            if self.use_hog:
                inputHOG = Input(batch_shape=(None, 2048),dtype='float32', name='input_HOG')
                normalizedHog = BatchNormalization(axis=-1)(inputHOG)
                outputImage = normalizedHog

            if self.use_lm and self.use_hog:
                outputImage = concatenate([normalizedHog, normalizedLandmarks])

            outputImage = Dense(
                units=32, activation=dense_activation)(outputImage)
            outputImage = BatchNormalization(axis=-1)(outputImage)

            outputImage = Dense(
                units=128, activation=dense_activation)(outputImage)
            outputImage = BatchNormalization(axis=-1)(outputImage)

        if (self.use_lm or self.use_hog) and self.use_cnn:
            concat_output = concatenate([outputCNN, outputImage])
        elif self.use_cnn:
            concat_output = outputCNN
        else:
            concat_output = outputImage

        output = Dense(units=128, activation=dense_activation)(concat_output)
        output = BatchNormalization(axis=-1)(output)

        output = Dense(units=512, activation=dense_activation)(output)
        output = BatchNormalization(axis=-1)(output)

        output = Dense(units=len(self.emotions), activation='softmax')(concat_output)

        # ========================== Model creating part ==========================
        inputs = []
        if self.use_cnn:
            inputs.append(input_image)
        if self.use_hog:
            inputs.append(inputHOG)
        if self.use_lm:
            inputs.append(inputLandmarks)

        model = Model(inputs=inputs, outputs=[output])

        model.compile(
            optimizer='adadelta',
            loss='categorical_crossentropy',
            metrics=['accuracy']
        )

        return model



#! /user/bin/env python3

import numpy as np
import cv2

def laplacian(img):
    laplacian = np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
    return cv2.filter2D(img, -1, laplacian)

# for removing salt and peper
def median(img, size=3):
    return cv2.medianBlur(img, size)

# for removing gausian
def fastNLMeans(img, h=15, templateWindowSize=7, searchWindowSize=21):
    return cv2.fastNlMeansDenoising(img, None, h, templateWindowSize, searchWindowSize)



import cv2
import dlib
import numpy as np

def faceTracking(frames, framesNumber):

    currentFaceID = -1
    frameCounter = -1

    nones = [None for i in range(framesNumber)]
    faces = []
    cords = []
    faceTrackers = {}

    # Loop over all faces and check if the area for this
    # face is the largest so far
    # We need to convert it to int here because of the
    # requirement of the dlib tracker. If we omit the cast to
    # int here, you will get cast errors since the detector
    # returns numpy.int32 and the tracker requires an int
    for (frame, croppedFaces) in frames:
        frameCounter += 1
        # frame = cv2.resize(frame, (320, 240))
        for ((_x1, _y1), (_x2, _y2)) in croppedFaces:
            x = int(_x1)
            y = int(_y1)
            w = int(abs(_x1-_x2))
            h = int(abs(_y1-_y2))

            # calculate the centerpoint
            x_bar = x + 0.5 * w
            y_bar = y + 0.5 * h

            matchedFid = False

            # Now loop over all the trackers and check if the
            # centerpoint of the face is within the box of a
            # tracker
            fidsToDelete = []
            for fid in faceTrackers.keys():
                trackingQuality = faceTrackers[fid].update(frame)
                if trackingQuality < 7:
                    fidsToDelete.append(fid)
                    continue

                tracked_position = faceTrackers[fid].get_position()

                t_x = int(tracked_position.left())
                t_y = int(tracked_position.top())
                t_w = int(tracked_position.width())
                t_h = int(tracked_position.height())

                # calculate the centerpoint
                t_x_bar = t_x + 0.5 * t_w
                t_y_bar = t_y + 0.5 * t_h

                # check if the centerpoint of the face is within the
                # rectangleof a tracker region. Also, the centerpoint
                # of the tracker region must be within the region
                # detected as a face. If both of these conditions hold
                # we have a match
                if ((t_x <= x_bar <= (t_x + t_w)) and
                        (t_y <= y_bar <= (t_y + t_h)) and
                        (x <= t_x_bar <= (x + w)) and
                        (y <= t_y_bar <= (y + h))):
                    matchedFid = True
                    faces[fid][frameCounter] = frame[y: (y + h), x: (x + w)]
                    cords[fid][frameCounter] = ((x, y), (x + w, y + h))
                    break

            for fid in fidsToDelete:
                faceTrackers.pop(fid, None)
            # If no matched fid, then we have to create a new tracker
            if not matchedFid:
                # Increase the currentFaceID counter
                currentFaceID += 1
                # Create and store the tracker
                tracker = dlib.correlation_tracker()
                tracker.start_track(frame, dlib.rectangle(x - 10, y - 20, x + w + 10, y + h + 20))
                faceTrackers[currentFaceID] = tracker
                faces.append(list(nones))
                cords.append(list(nones))
                faces[currentFaceID][frameCounter] = frame[y: (y + h), x: (x + w)]
                cords[currentFaceID][frameCounter] = ((x, y), (x + w, y + h))

    return faces, cords




from threading import Thread, get_ident, Condition
from queue import Queue
import time

import cv2
from imutils.video import VideoStream

from Cerebro.interface import process_image as pi

frame_data = []
def detect_stream_emotions(skip=10):
    frame_counter = 0
    vs = cv2.VideoCapture(0)

    worker_thread = WorkerThread()
    worker_thread.start()

    while True:
        success, frame = vs.read()
        if not success: continue
        
        frame_counter += 1

        if (frame_counter % skip) == 0:
            worker_thread.add_task(frame)
            frame_counrer = 0

        if len(frame_data) > 0:
            frame = pi.mark_faces_emotions(frame, None, frame_data)

        cv2.imshow("Frame", frame)

        if cv2.waitKey(1) & 0xFF == ord("q"): break
    
    vs.release()
    cv2.destroyAllWindows()
    worker_thread.kill()
    worker_thread.join()

class WorkerThread(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.condition = Condition()
        self.saved_frame = None
        self.should_close = False
    
    def run(self):
        global frame_data
        while True:
            # get saved frame if any and clear the saved value
            self.condition.acquire()
            try:
                while type(self.saved_frame) == type(None): 
                    if self.should_close: return
                    self.condition.wait()
                
                frame = self.saved_frame
                self.saved_frame = None
            finally:
                self.condition.release()
            
            if type(frame) == type(-1) and frame == -1: break
            frame_data = pi.extract_faces_emotions(frame, 'haar')
    
    def add_task(self, frame):
        self.condition.acquire()
        try:
            self.saved_frame = frame
            self.condition.notify()
        finally:
            self.condition.release()

    def kill(self):
        self.condition.acquire()
        try:
            self.should_close = True
            self.condition.notify()
        finally:
            self.condition.release()



#! /user/bin/env python3

import cv2
from skimage.feature import hog
from imutils import face_utils
import dlib
import os
import Cerebro

landmarks_model_path = os.path.join(
	Cerebro.__cwd__,
	'saved-models/face-landmarks/shape_predictor_68_face_landmarks.dat'
)
if not os.path.isfile(landmarks_model_path):
	raise Exception("shape_predictor_68_face_landmarks.dat doesn't exist.")

__dlib_landmark_predictor = dlib.shape_predictor(landmarks_model_path)

def sk_get_hog(img, orientations=8, pixels_per_cell=(12, 12)):
	"""
	img : 48 * 48 gray scale image
	"""
	features, hog_image = hog(
		img,
		orientations=orientations,
		pixels_per_cell=pixels_per_cell,
		cells_per_block=(4, 4),
		visualize=True,
		transform_sqrt=False,
		feature_vector=True,
		multichannel=False,
		block_norm='L2'
	)

	return features

def get_face_landmarks(img):
	rect = dlib.rectangle(0, 0, img.shape[0] - 1, img.shape[1] -1)
	shape = __dlib_landmark_predictor(img, rect)
	shape = face_utils.shape_to_np(shape)
	return shape

def hog_with_sliding_window(img, window_step=6, orientations=8, pixels_per_cell=(12, 12), cells_per_block=(1, 1), block_norm='L2-Hys'):
	"""
	window_step = 6
	img 48*48 gray scale
	-------------------
	orientations = 8
	pixels_per_cell = (12, 12) >> output length : 1152
	pixels_per_cell = (8, 8) >> output length : 2592 <<<<<
	-------------------
	orientations = 12
	pixels_per_cell = (12, 12) >>output length : 1728
	pixels_per_cell = (12, 12) >>output length : 3888
	any change in the default parameters would change the output length
	"""
	window_size = 24
	hog_windows =[]
	for y in range(0, img.shape[0], window_step):
		for x in range(0, img.shape[1], window_step):
			window = img[y:y+window_size, x:x+window_size]
			hog_windows.extend(hog(
			window,
			orientations=orientations,
			pixels_per_cell=pixels_per_cell,
			cells_per_block=cells_per_block,
			block_norm=block_norm,
			visualize=False
			))
	return hog_windows



from Cerebro.image.face_detector.__cascade_face_detector import cascade_FaceDetector

__lbp_facedetector = cascade_FaceDetector("lbp")

def get_faces(img):
    return __lbp_facedetector.get_faces(img)

def is_one_face(img):
    return __lbp_facedetector.is_one_face(img)



#! /user/bin/env python3

import argparse

from Cerebro.model.emotions_model import *
from Cerebro.model.reader import read_testing, read_training, get_emotions, get_reduced_emotions, set_dataset, available_datasets

def main():
	parser = argparse.ArgumentParser(description='')

	parser.add_argument('-g', action='store_true', help='Use HOG for training')
	parser.add_argument('-m', action='store_true', help='Use LM (land marks) for training')
	parser.add_argument('-c', action='store_true', help='Use CNN for training')

	parser.add_argument('-t', action='store_true', help='Force train model')
	parser.add_argument('-n', action='store_true', help='Create new model, reset old weights (if any)')

	parser.add_argument('--tr-count', metavar='int', type=int, default=-1, help='Set the count of the training data')
	parser.add_argument('--ts-count', metavar='int', type=int, default=-1, help='Set the count of the testing data')
	
	parser.add_argument('-d', type=str, required=True, choices=available_datasets, help='Dataset to use')

	args = parser.parse_args()
	must_train = args.t

	set_dataset(args.d)

	print('Reading data...')
	x_test, y_test = read_testing(args.ts_count)
	x_train, y_train = read_training(args.tr_count)

	model = EmotionsModel(
				use_lm=args.m,
				use_hog=args.g,
				use_cnn=args.c,
				create_new=args.n,
				emotions=get_emotions(),
				reduced_emotions=get_reduced_emotions(),
				verbose=True,
			)

	print('\nCreated Model.')

	if not model.is_trained or args.t:
		print('Training...')
		history = model.fit(x_train, y_train)
		# print(history.history['val_acc'])

	print('\nTesting...')
	result = model.test(x_test, y_test)
	print("Result:", result)

if __name__ == '__main__': main()


from Cerebro.image.face_detector.__cascade_face_detector import cascade_FaceDetector

__haar_facedetector = cascade_FaceDetector("haar")

def get_faces(img):
    return __haar_facedetector.get_faces(img)

def is_one_face(img):
    return __haar_facedetector.is_one_face(img)




import cv2
import threading

from Cerebro.model import emotions_model as model
from Cerebro.image.face_detector import detect_dlib
from Cerebro.image.face_detector import detect_haar
from Cerebro.image.face_detector import detect_lbp

emotionsModel = model.EmotionsModel()

def extract_faces_emotions(image, detector_type = 'dlib'):

    items= []
    if detector_type == 'dlib':
        detector = detect_dlib
    elif detector_type =='haar':
        detector = detect_haar
    elif detector_type =='lbp':
        detector = detect_lbp
    else :
        raise Exception("invalid detector")

    faces = detector.get_faces(image)

    # emotions_count = len(set(emotions_map))
    # emotionsModel = model.EmotionsModel()
    for i in range (len(faces)) :
        item = []
        item.append(faces[i][0]) #face
        item.append(faces[i][1]) #corner coordinates
        if faces[i][0].shape[0] == 0 or faces[i][0].shape[1] == 0: continue
        emotion = emotionsModel.predict(faces[i][0])
        item.append(emotion)
        items.append(item)
    return items

def mark_faces_emotions(image, detector_type = 'dlib', extracted_faces_emotions = None):
    """
    takes image(numpy array) and returns the same image with marked faces and emotions.
    you can control detector type passing one of the available detectors (default:'dlib', 'haar', 'lbp').
    in case of calling extracted_faces_emotions to use it, you shold pass its return here in extracted_faces_emotions parameter.
    """
    font = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = 1
    color = (72, 1, 68)
    offset_x = 5
    offset_y = 0

    # if detector_type >> None : don't detect
    if detector_type != None and extracted_faces_emotions == None:
        extracted_faces_emotions = extract_faces_emotions(image, detector_type)

    """
    extracted_faces_emotions[i] >> item (face, corners(topright,bottomleft), emotion)
    extracted_faces_emotions[i][1] >> corners
    extracted_faces_emotions[i][1][0] >> topright
    extracted_faces_emotions[i][1][0][0] >> toprightX
    extracted_faces_emotions[i][1][0][1] >> toprightY
    extracted_faces_emotions[i][2] >> emotion
    """

    if extracted_faces_emotions == None: return image

    for item in extracted_faces_emotions:
        tmp = (item[1][0][0] + offset_x, item[1][0][1] - offset_y)

        #text background
        size = cv2.getTextSize(item[2], font, fontScale=font_scale, thickness=1)[0]
        box_coords = (tmp, (tmp[0] + size[0] - 2, tmp[1] - size[1] - 2))
        cv2.rectangle(image, box_coords[0], box_coords[1], color, cv2.FILLED)

        #selected face box
        image = cv2.rectangle(
            image,
            item[1][0],
            item[1][1],
            color,
            2,
        )

        #text
        image = cv2.putText(
            image,
            item[2],
            tmp,
            font,
            font_scale,
            (255, 255, 255),
            1,
        )

    return image



from distutils.core import setup
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = 'CEREBRO8',
    version = '0.0.1',
    license='gpl-3.0',
    description = 'Recognizing facial expressions from images, videos and real-time stream',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author = 'AmrSaber, WafaaIsmail, MohamedAhmed, SalmaSayed, MuhammadAref, ManarArabi',
    author_email = 'amr.m.saber.mail@gmail.com, wafaaismail595@gmail.com, mohamedmaim97@gmail.com, Salmasayed797@gmail.com, muhammad.aref224@gmail.com, manar.araby.ma@gmail.com',
    url = 'https://github.com/AmrSaber/Cerebro',
    download_url = 'https://github.com/AmrSaber/Cerebro/archive/v1.0.tar.gz',
    keywords = ['emotions', 'expressions', 'real-time stream'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        ],
    install_requires=[
#            'opencv-python',
            'opencv-contrib-python==3.4.4.19',
            'numpy',
            'moviepy',
            'ffmpeg-python',
            'imutils',
            'argparse',
            'keras',
            'pathlib',
            'matplotlib',
            'scikit-image',
            'dlib',
            'Theano'
            ]
)
print(setuptools.find_packages())


import cv2
import Cerebro
import os 
# haarcascade_frontalface_alt.xml scale-factor = 1.00849, min-neighbours = 5
# scale factor of 1.03 - 1.05 for haar seems to give good results

class cascade_FaceDetector:
	def __init__(self, classifier_type):
		self.min_neighbors = 5
		if classifier_type == 'haar':
			path = os.path.join(Cerebro.__cwd__,
					'saved-models/face-detection/haarcascade_frontalface_alt.xml'
					)
			if not os.path.isfile(path):
				raise Exception("haarcascade_frontalface_alt.xml doesn't exist.")
			self.classifier = cv2.CascadeClassifier(path)
			self.scale_factor = 1.0303035
		elif classifier_type == 'lbp':
			path = os.path.join(Cerebro.__cwd__,
				'saved-models/face-detection/lbpcascade_frontalface_improved.xml')
			if not os.path.isfile(path):
				raise Exception("lbpcascade_frontalface_improved.xml doesn't exist.")
			self.classifier = cv2.CascadeClassifier(path)
			self.scale_factor = 1.2
		else:
			raise Exception("unknown classifier type")


	def get_faces(self, img):
		"""
		this function is supposed to extract faces from image, set them in required format,
		enhance them, pack them with their respective boxes in original image and return them
		:param img:np.mat the image with faces in it
		:returns res:list [ .., (face, box), ...]
		where each face is an image for the detected face
		and each box is a tuple of two points the upper-left-corner and lower-right-corner
		"""
		boxes = self.__get_faces_locations(img)
		res = []
		for box in boxes:
			face = self.__extract_face(img, box)
			# face = self.__enhance_face(face)
			res.append((face, box))
		return res

	def is_one_face(self, img):
		"""
		this function checks if the given image contains EXACTLY 1 face
		:param img:
		:return: True if only 1 face is found, False otherwise
		"""
		return len(self.__get_faces_locations(img)) == 1

	# private functions
	def __get_faces_locations(self, img):
		"""
		this function applies the haar classifier to the image to detect the faces in it
		:param img:np.mat
		:return:boxes:list of boxes that contain faces in the image
		where each box contains a tuple of two points
		the upper left corner and the lower right corner
		"""
		faces = self.classifier.detectMultiScale(
			img,
			scaleFactor=self.scale_factor,
			minNeighbors=self.min_neighbors
		)
		boxes = []
		for face in faces:
			p1 = (face[0], face[1])
			p2 = (face[2] + face[0], face[3] + face[1])
			boxes.append((p1, p2))
		return boxes

	def __extract_face(self, img, box):
		"""
		this function should extract the faces in the given locations in the given image
		:param img:np.mat the image with faces in it
		:param boxes:list it contains the locations for faces in the image
		:return:faces:list of np.mat images of faces in the given image
		"""
		p1, p2 = box
		face = img[p1[1]:p2[1], p1[0]:p2[0]]
		return face

	@staticmethod
	def __display(img, boxes):
		for box in boxes:
			cv2.rectangle(img, box[0], box[1], (0, 255, 0), 1)
		cv2.imshow("detected faces", img)
		cv2.waitKey(0)




from Cerebro.interface import process_video as pi
import cv2 
import Cerebro as c

def main():
	c.set_cwd("C:/Users/manar/Desktop/Cerebro/Cerebro")
	pi.detect_video_emotions_with_tracking("c.mp4","c_out_tracking.mp4",50, verbose=True)
	
	
if __name__== "__main__":
  main()